"""
Calculations provided by aiida_aurora.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""

from .cycler import BatteryCyclerExperiment
from .fake import BatteryFakeExperiment

__all__ = [
    'BatteryCyclerExperiment',
    'BatteryFakeExperiment',
]



"""
A dummy experiment specifications class.
"""

import json

import yaml

from aiida.orm import Dict

from aiida_aurora.schemas.cycling import ElectroChemSequence as ElectroChemSequenceSchema


class CyclingSpecsData(Dict):  # pylint: disable=too-many-ancestors
    """
    An experiment specification object.

    This class represents the specifications used in an experiment.
    """

    # "pydantic" schema  to add automatic validation
    schema = ElectroChemSequenceSchema

    # pylint: disable=redefined-builtin
    def __init__(self, dict=None, **kwargs):
        """
        Constructor for the data class

        Usage: ``DummyExperimentSpecs(dict{...})``

        :param parameters_dict: dictionary with battery specifications
        :param type parameters_dict: dict
        """
        dict = self.validate(dict)
        super().__init__(dict=dict, **kwargs)
        # if not self.label:
        #     self.label = self.dict['metadata'].get('name')

    def validate(self, parameters_dict):  # pylint: disable=no-self-use
        """Validate command line options.

        Uses the voluptuous package for validation. Find out about allowed keys using::

            print(DummyExperimentSpecs.schema)

        :param parameters_dict: dictionary with battery specifications
        :param type parameters_dict: dict
        :returns: validated dictionary
        """
        return ElectroChemSequenceSchema(**parameters_dict).dict()

    def get_json(self):
        """Get a JSON file containing the DummyExperimentSpecs specs."""

        # this can be customized to fit the desired format
        object_to_be_serialized = self.get_dict()
        return json.dumps(object_to_be_serialized)

    def get_yaml(self):
        """Get a YAML file containing the BatterySampleData specs."""

        # this can be customized to fit the desired format
        object_to_be_serialized = {"method": self.get_dict()}
        return yaml.dump(object_to_be_serialized)

    def __str__(self):
        """String representation of node.

        Append values of dictionary to usual representation. E.g.::

            uuid: b416cbee-24e8-47a8-8c11-6d668770158b (pk: 590)
            {'ignore-case': True}

        """
        string = super().__str__()
        string += "\n" + str(self.get_dict())
        return string



"""
Data types provided by plugin

Register data types via the "aiida.data" entry point in setup.json.
"""

from .battery import BatterySampleData, BatteryStateData
from .control import TomatoSettingsData
from .experiment import CyclingSpecsData

__all__ = [
    'BatterySampleData',
    'BatteryStateData',
    'TomatoSettingsData',
    'CyclingSpecsData',
]



"""
Plugin for the Tomato scheduler.
"""
import datetime
import re

import yaml

from aiida.common import exceptions
from aiida.common.escaping import escape_for_bash
from aiida.common.extendeddicts import AttributeDict
from aiida.schedulers import Scheduler, SchedulerError
from aiida.schedulers.datastructures import JobInfo, JobResource, JobState

# There is no "job owner" in tomato.

# Mapping of Tomato states to AiiDA `JobState`s
#
# The following statuses are defined in tomato:
#  - q   job is queued. Jobs shouldn't stay in q too long as that indicates there
#        is no pipeline that can process the payload.
#  - qw  job is queued and a matching pipeline has been found, but it is either busy,
#        not ready, or without the correct sample
#  - r   job is running
#  - c   job has completed successfully
#  - ce  job has completed with an error - output data not guaranteed, might be present in the job folder.
#  - cd  job has been cancelled - output data should be available as specified in the yamlfile

_MAP_STATUS_TOMATO = {
    "q": JobState.QUEUED,  # JobState.QUEUED_HELD ?
    "qw": JobState.QUEUED,
    "r": JobState.RUNNING,
    "c": JobState.DONE,
    "ce": JobState.DONE,
    "cd": JobState.DONE,
}

_MAP_ANNOTATION_TOMATO = {
    "q": "Queued",
    "qw": "Queued, matching pipeline found",
    "r": "Running",
    "c": "Completed successfully",
    "ce": "Completed with error",
    "cd": "Cancelled",
}


class TomatoResource(JobResource):
    """Class for Tomato job resources."""

    @classmethod
    def validate_resources(cls, **kwargs):
        """Validate the resources against the job resource class of this scheduler.

        :param kwargs: dictionary of values to define the job resources
        :raises ValueError: if the resources are invalid or incomplete
        :return: optional tuple of parsed resource settings
        """
        resources = AttributeDict()
        # *Essentially ignore all job resources!*
        return resources

    def __init__(self, **kwargs):
        """Initialize the job resources from the passed arguments.

        :raises ValueError: if the resources are invalid or incomplete
        """
        resources = self.validate_resources(**kwargs)
        super().__init__(resources)

    @classmethod
    def accepts_default_mpiprocs_per_machine(cls):
        """Return True if this subclass accepts a `default_mpiprocs_per_machine` key, False otherwise."""
        return False

    def get_tot_num_mpiprocs(self):
        """Return the total number of cpus of this job resource."""
        return 1


class TomatoScheduler(Scheduler):
    """
    Support for the Tomato scheduler (https://github.com/dgbowl/tomato)
    Supports tomato version 0.2a1
    """

    _logger = Scheduler._logger.getChild("tomato")

    # Query only by list of jobs and not by user
    _features = {
        "can_query_by_user": False,
    }

    # The class to be used for the job resource.
    _job_resource_class = TomatoResource

    _map_status = _MAP_STATUS_TOMATO

    # the command used to submit the script
    _shell_cmd = ""

    # the scheduler command
    # (NOTE: if applicable, you should configure the computer to load the appropriate virtual environment,
    # in order to make this command available)
    KETCHUP = "ketchup"

    def _get_joblist_command(self, jobs=None, user=None):
        """The command to report full information on existing jobs.

        :return: a string of the command to be executed to determine the active jobs.
        """

        if user:
            raise exceptions.FeatureNotAvailable("Cannot query by user")

        if jobs:
            if isinstance(jobs, str):
                command = f"{self.KETCHUP} status {escape_for_bash(jobs)}"
            else:
                command = f"{self.KETCHUP} status "
                try:
                    command += " ".join(f"{j}" for j in jobs)
                except TypeError as e:
                    raise TypeError(
                        "If provided, the 'jobs' variable must be a string or an iterable of strings"
                    ) from e
        else:
            command = f"{self.KETCHUP} status queue -v"

        self._logger.debug(f"ketchup command: {command}")
        return command

    def _get_detailed_job_info_command(self, job_id):
        """Return the command to run to get the detailed information on a job,
        even after the job has finished.

        The output text is just retrieved, and returned for logging purposes.
        """
        return f"{self.KETCHUP} status {escape_for_bash(job_id)}"

    def _get_submit_script_header(self, job_tmpl):
        """Return the submit script final part, using the parameters from the job template.

        :param job_tmpl: a ``JobTemplate`` instance with relevant parameters set.
        """
        # set the `_shell_cmd` class attribute from the job template
        # FIXME shell_type used to be job_tmpl.shell_type, which was added to
        # aiida-core's `JobTemplate` class by Loris. Awaiting resolution.
        shell_type = "powershell"
        # shell_type = "bash"  # uncomment when debugging on Linux
        TomatoScheduler._shell_cmd = shell_type
        self._logger.debug(f"_get_submit_script_header: _shell_cmd: {self._shell_cmd}")

        import string

        if job_tmpl.job_name:
            # I leave only letters, numbers, dots, dashes and underscores
            # Note: I don't compile the regexp, I am going to use it only once
            job_title = re.sub(r"[^a-zA-Z0-9_.-]+", "", job_tmpl.job_name)

            # prepend a 'j' (for 'job') before the string if the string
            # is now empty or does not start with a valid character
            if not job_title or (job_title[0] not in string.ascii_letters + string.digits):
                job_title = f"j{job_title}"

            # Truncate to the first 128 characters
            # Nothing is done if the string is shorter.
            job_title = job_title[:128]
            if shell_type == "powershell":
                header = f"$JOB_TITLE='{job_title}'"
            else:
                header = f"JOB_TITLE='{job_title}'"
        else:
            header = ""

        return header

    def _get_submit_command(self, submit_script):
        """Return the string to execute to submit a given script.

        .. warning:: the `submit_script` should already have been bash-escaped

        :param submit_script: the path of the submit script relative to the working directory.
        :return: the string to execute to submit a given script.
        """
        # Similarly to the 'direct' scheduler, we submit a bash script that actually executes
        # `ketchup submit {payload}`
        # the output of it is parsed *immediately* by _parse_submit_output
        submit_command = f"{self._shell_cmd} {submit_script}"

        self._logger.info(f"submitting with: {submit_command}")

        return submit_command

    def _get_run_line(self, codes_info, codes_run_mode):
        """Return a string with the line to execute a specific code with specific arguments.

        :parameter codes_info: a list of `aiida.common.datastructures.CodeInfo` objects. Each contains the information
            needed to run the code. I.e. `cmdline_params`, `stdin_name`, `stdout_name`, `stderr_name`, `join_files`. See
            the documentation of `JobTemplate` and `CodeInfo`.
        :parameter codes_run_mode: instance of `aiida.common.datastructures.CodeRunMode` contains the information on how
            to launch the multiple codes.
        :return: string with format: [executable] [args] {[ < stdin ]} {[ < stdout ]} {[2>&1 | 2> stderr]}

        Tomato: the only customization with respect to the base-class `_get_run_line` method consists in not using
        `escape_for_bash`, because we want to make use of environmental variables in the run line.
        """
        from aiida.common.datastructures import CodeRunMode

        list_of_runlines = []

        for code_info in codes_info:
            command_to_exec_list = []
            for arg in code_info.cmdline_params:
                command_to_exec_list.append(arg)
                # command_to_exec_list.append(escape_for_bash(arg))
            command_to_exec = " ".join(command_to_exec_list)

            stdin_str = (f"< {escape_for_bash(code_info.stdin_name)}" if code_info.stdin_name else "")
            stdout_str = (f"> {escape_for_bash(code_info.stdout_name)}" if code_info.stdout_name else "")

            join_files = code_info.join_files
            if join_files:
                stderr_str = "2>&1"
            else:
                stderr_str = (f"2> {escape_for_bash(code_info.stderr_name)}" if code_info.stderr_name else "")

            output_string = f"{command_to_exec} {stdin_str} {stdout_str} {stderr_str}"

            list_of_runlines.append(output_string)

        self.logger.debug(f"_get_run_line output: {list_of_runlines}")

        if codes_run_mode == CodeRunMode.PARALLEL:
            list_of_runlines.append("wait\n")
            return " &\n\n".join(list_of_runlines)

        if codes_run_mode == CodeRunMode.SERIAL:
            return "\n\n".join(list_of_runlines)

        raise NotImplementedError("Unrecognized code run mode")

    def _parse_joblist_output(self, retval, stdout, stderr):
        """Parse the joblist output as returned by executing the command returned by `_get_joblist_command` method.

        :return: list of `JobInfo` objects, one of each job each with at least its default params implemented.
        """
        if retval != 0:
            raise SchedulerError(
                f"""ketchup returned exit code {retval} (_parse_joblist_output function)"""
                f"""stdout='{stdout.strip()}'"""
                f"""stderr='{stderr.strip()}'"""
            )
        if stderr.strip():
            self.logger.warning(
                f"ketchup returned exit code 0 (_parse_joblist_output function) but non-empty stderr='{stderr.strip()}'"
            )

        # remove all empty lines and lines containing 'ERROR'
        jobdata_raw = "\n".join(l for l in stdout.splitlines() if l and "ERROR" not in l)

        def convert_datetime(dt):
            if isinstance(dt, datetime.datetime):
                return dt
            try:
                return datetime.datetime.fromisoformat(dt)
            except Exception:
                return None

        # Create dictionary and parse specific fields
        job_list = []

        if "===========================" in jobdata_raw:
            # the command was 'ketchup status queue'
            jobdata_raw = jobdata_raw.splitlines()
            if len(jobdata_raw) > 2:
                for line in jobdata_raw[2:]:
                    job = line.split()
                    this_job = JobInfo()
                    this_job.job_id = job[0]
                    this_job.title = job[1]

                    try:
                        this_job.job_state = _MAP_STATUS_TOMATO[job[2]]
                        this_job.annotation = _MAP_ANNOTATION_TOMATO[job[2]]
                    except KeyError:
                        self.logger.warning(f"Unrecognized job_state '{job[2]}' for job id {this_job.job_id}")
                        this_job.job_state = JobState.UNDETERMINED

                    if len(job) == 3:
                        this_job.pipeline = None
                    elif len(job) == 4:
                        this_job.pipeline = job[3]
                    else:
                        raise ValueError(f"More than 4 columns returned by ketchup status queue\n{job}")

                    # Everything goes here anyway for debugging purposes
                    this_job.raw_data = job

                    # I append to the list of jobs to return
                    job_list.append(this_job)
            else:
                pass  # there are no jobs in the queue

        else:
            # the command was 'ketchup status {jobid} ...'
            # the output is yaml-formatted
            jobdata_parsed = yaml.full_load(jobdata_raw)

            for this_job_dict in jobdata_parsed:
                this_job = JobInfo()
                this_job.job_id = str(this_job_dict["jobid"])
                this_job.title = this_job_dict["jobname"]
                try:
                    this_job_status = this_job_dict["status"]
                    this_job.job_state = _MAP_STATUS_TOMATO[this_job_status]
                    this_job.annotation = _MAP_ANNOTATION_TOMATO[this_job_status]
                except KeyError:
                    self.logger.warning(f"Unrecognized job_state '{this_job_status}' for job id {this_job.job_id}")
                    this_job.job_state = JobState.UNDETERMINED

                # yaml parses iso-format datetime strings automatically:
                this_job.submission_time = convert_datetime(this_job_dict.get("submitted"))
                this_job.dispatch_time = convert_datetime(this_job_dict.get("executed"))
                this_job.finish_time = convert_datetime(this_job_dict.get("completed"))
                this_job.allocated_machines = this_job_dict.get("pipeline")
                # ignored: this_job_dict.get("pid")

                job_list.append(this_job)  # append last job

        return job_list

    def _parse_submit_output(self, retval, stdout, stderr):
        """Parse the output of the submit command returned by calling the `_get_submit_command` command.

        :return: a string with the job ID.
        """
        if retval != 0:
            self._logger.error(f"Error in _parse_submit_output: retval={retval}; stdout={stdout}; stderr={stderr}")
            raise SchedulerError(f"Error during submission, retval={retval}; stdout={stdout}; stderr={stderr}")

        if stderr.strip():
            self._logger.warning(f"in _parse_submit_output there was some text in stderr: {stderr}")

        # I check for the jobid in the output
        stdout_dict = yaml.full_load(stdout)
        if "jobid" in stdout_dict:
            self._logger.debug(f"The submitted jobid is {stdout_dict['jobid']}")
            # HACK did not need to str-cast prior to aiida 2.x upgrade
            return str(stdout_dict["jobid"])

        # If I am here, no jobid was found
        self.logger.error(f"in _parse_submit_output: unable to find the job id: {stdout}")
        raise SchedulerError(
            "Error during submission, could not retrieve the jobID from ketchup output; see log for more info."
        )

    def _get_kill_command(self, jobid):
        """Return the command to kill the job with specified jobid."""

        kill_command = f"{self.KETCHUP} cancel {jobid}"

        self._logger.info(f"killing job {jobid}: {kill_command}")

        return kill_command

    def _parse_kill_output(self, retval, stdout, stderr):
        """Parse the output of the kill command.

        :return: True if everything seems ok, False otherwise.
        """
        if retval != 0:
            self._logger.error(f"Error in _parse_kill_output: retval={retval}; stdout={stdout}; stderr={stderr}")
            return False

        if stderr.strip():
            self._logger.warning(f"in _parse_kill_output there was some text in stderr: {stderr}")

        if stdout.strip():
            self._logger.warning(f"in _parse_kill_output there was some text in stdout: {stdout}")

        return True

    def parse_output(self, detailed_job_info, stdout, stderr):
        """Parse the output of the scheduler.

        :param detailed_job_info: dictionary with the output returned by the `Scheduler.get_detailed_job_info` command.
            This should contain the keys `retval`, `stdout` and `stderr` corresponding to the return value, stdout and
            stderr returned by the accounting command executed for a specific job id.
        :param stdout: string with the output written by the scheduler to stdout
        :param stderr: string with the output written by the scheduler to stderr
        :return: None or an instance of `aiida.engine.processes.exit_code.ExitCode`
        :raises TypeError or ValueError: if the passed arguments have incorrect type or value
        """
        raise NotImplementedError()



from __future__ import annotations

import json

from pandas import DataFrame
from pandas.io.formats.style import Styler

from aiida.orm import CalcJobNode, QueryBuilder, RemoteData, SinglefileData

from aiida_aurora.data import BatterySampleData
from aiida_aurora.utils.parsers import get_data_from_raw, get_data_from_results


def cycling_analysis(node: CalcJobNode) -> tuple[dict, str, DataFrame]:
    """Perform post-processing of cycling experiments results.

    Used by the frontend Aurora app for plotting.

    The analysis report attached to each plot series includes:
    - details of the associated monitor, if one was assigned
    - summary/statistics of the post-processed data

    Parameters
    ----------
    `node` : `CalcJobNode`
        The calculation `node`.

    Returns
    -------
    `tuple[dict, str, DataFrame]`
        Post-processed data and an analysis report.

    Raises
    ------
    `TypeError`
        If `node` is not a `BatteryCyclerExperiment`.
    """

    if node.process_type != "aiida.calculations:aurora.cycler":
        raise TypeError("`node` is not a `BatteryCyclerExperiment`")

    log = f"CalcJob:   <{node.pk}> '{node.label}'\n"

    sample: BatterySampleData = node.inputs.battery_sample
    log += f"Sample:    {sample.label}\n"

    log += "Monitored: "

    if monitors := get_monitors(node):
        log += "True"
        log += add_monitor_details(monitors)
    else:
        log += "False"

    try:
        data, warning, raw = process_data(node)
    except Exception as err:
        data, warning, raw = {}, f"*** ERROR ***\n\n{str(err)}", None

    return (data, f"{log}\n{warning}", raw)


def get_monitors(node: CalcJobNode) -> dict[str, dict]:
    """Fetch the monitor dictionary.

    The function is backwards compatible, capable of fetching the
    dictionary of monitors defined in the AiiDA 2.x monitoring
    feature, or, if the calculation was submitted prior to the
    AiiDA 2.x update, fetch the associated monitor calcjob and
    prepare the dictionary in accordance with the new format.

    Parameters
    ----------
    `node` : `CalcJobNode`
        The calculation node.

    Returns
    -------
    `dict[str, dict]`
        A dictionary of monitors.
    """

    if "monitors" in node.inputs:
        return {k: dict(v) for k, v in dict(node.inputs.monitors).items()}

    # BACKWARDS COMPATABILITY
    # job submitted prior to AiiDA 2.x upgrade - fetch monitor calcjob
    monitor = get_node_monitor_calcjob(node)
    return convert_to_new_monitor_format(monitor) if monitor else {}


def get_node_monitor_calcjob(node: CalcJobNode) -> CalcJobNode | None:
    """Fetch the monitor calcjob associated with the calculation
    `node`.

    Uses the `QueryBuilder` to query for a monitor calcjob with
    a `RemoteData` node associated with the calculation node.

    Parameters
    ----------
    `node` : `CalcJobNode`
        The calculation node.

    Returns
    -------
    `CalcJobNode | None`
        The associated monitor calcjob node, `None` if not found.
    """

    if "remote_folder" not in node.outputs:
        return None

    remote_folder: RemoteData = node.outputs.remote_folder

    qb = QueryBuilder()

    qb.append(
        RemoteData,
        filters={
            'uuid': remote_folder.uuid,
        },
        tag='remote_folder',
    ).append(
        CalcJobNode,
        with_incoming='remote_folder',
        edge_filters={
            'label': 'monitor_folder'
        },
        project=['*', 'id'],
        tag='monitor',
    ).order_by({
        'monitor': {
            'id': 'desc'
        },
    })

    results = qb.first() if qb.count() else None
    return results[0] if results else None


def convert_to_new_monitor_format(monitor: CalcJobNode) -> dict[str, dict]:
    """Convert monitor calcjob attributes to AiiDA 2.x format.

    For more details, see

    https://aiida.readthedocs.io/projects/aiida-core/en/latest/howto/run_codes.html#how-to-monitor-and-prematurely-stop-a-calculation

    Parameters
    ----------
    `monitor` : `CalcJobNode`
        The monitor calcjob `node`.

    Returns
    -------
    `dict[str, dict]`
        The formatted monitor dictionary.
    """

    protocols: dict = monitor.inputs.monitor_protocols
    params: dict = protocols["monitor1"].attributes

    settings: dict = params.get("options", {})
    sources: dict = params.get("sources", {})
    extra: dict = sources.get("output", {})

    refresh_rate = extra.get("refresh_rate", 600)
    filename = extra.get("filepath", "snapshot.json")

    threshold = settings.get("threshold", 0.8)
    check_type = settings.get("check_type", "discharge_capacity")
    consecutive_cycles = settings.get("consecutive_cycles", 2)

    return {
        "capacity": {
            "entry_point": f"aiida-calcmonitor plugin <pk={monitor.pk}>",
            "minimum_poll_interval": refresh_rate,
            "kwargs": {
                "filename": filename,
                "settings": {
                    "threshold": threshold,
                    "check_type": check_type,
                    "consecutive_cycles": consecutive_cycles,
                }
            }
        }
    }


def add_monitor_details(monitors: dict[str, dict]) -> str:
    """Return monitor details.

    Details include the following:
    - monitor label
    - AiiDA entry point of the monitor function/calcjob
    - refresh (polling) rate
    - source file to be polled
    - monitor settings

    Parameters
    ----------
    `monitors` : `dict[str, dict]`
        A dictionary of monitors.

    Returns
    -------
    `str`
        The monitor details to be added to the analysis report.
    """

    details = "" if monitors else "\nWARNING: No monitors found\n"

    for label, params in monitors.items():
        details += f"\nMonitor:              {label}\n"
        refresh_rate = params.get("minimum_poll_interval", 600)
        details += f"  Interval (s):       {refresh_rate}\n"

        if "kwargs" in params:
            kwargs: dict = params["kwargs"]
            source_file = kwargs.get("filename", "snapshot.json")
            settings: dict = kwargs.get("settings", {})
            details += add_monitor_settings(source_file, settings)

    return details


def add_monitor_settings(
    source_file: str,
    settings: dict,
) -> str:
    """Return specific monitor settings details.

    NOTE: Setting keys are sentence-cased.

    Parameters
    ----------
    `source_file` : `str`
        The polled source file.
    `settings` : `dict`
        The monitor settings.

    Returns
    -------
    `str`
        Details of the monitor settings.
    """

    _settings = f"  Source file:        {source_file}\n"

    check_type = settings.pop("check_type", "discharge_capacity")
    _settings += f"  Check type:         {check_type}\n"

    key: str
    for key, value in settings.items():
        key = key.replace("_", " ").capitalize() + ":"
        _settings += f"  {key:19s} {value}\n"

    return _settings


def process_data(node: CalcJobNode) -> tuple[dict, str, Styler | str]:
    """Analyze the results of the cycling experiment.

    The analysis is performed on the results `ArrayNode`, if one
    was prepared by AiiDA upon a successful run. If not, in the
    case the job was terminated prematurely, the function will
    attempt to analyze (in order) the raw (non-parsed) results,
    the retrieved results file, or if none was retrieved, the
    snapshot fetched directly from the remote machine.

    Parameters
    ----------
    `node` : `CalcJobNode`
        The calculation `node`.

    Returns
    -------
    `tuple[dict, str, Styler | str]`
        Post-processed data, warning, and analysis | error message.
    """

    if node.process_state and "finished" not in node.process_state.value:
        return {}, f"Job terminated with message '{node.process_status}'", ""

    warning = ""

    if node.exit_status:
        warning += "WARNING: "
        generic = "job killed by monitor"
        warning += f"{node.exit_message}" if node.exit_message else generic
        warning += "\n\n"

    if "results" in node.outputs:
        data = get_data_from_results(node.outputs.results)
    elif "raw_data" in node.outputs:
        data = get_data_from_file(node.outputs.raw_data)
    elif "retrieved" in node.outputs:
        data = get_data_from_file(node.outputs.retrieved)
    elif "remote_folder" in node.outputs:
        data = get_data_from_remote(node.outputs.remote_folder)
    else:
        data = {}

    return data, warning, add_analysis(data)


def get_data_from_file(source: SinglefileData) -> dict:
    """Return source file as a post-processed dictionary.

    NOTE: assumes file is of JSON format.

    Parameters
    ----------
    `source` : `SinglefileData`
        The node of the associated retrieved results file.

    Returns
    -------
    `dict`
        The post-processed data dictionary.
    """
    if "results.json" in source.base.repository.list_object_names():
        file = source.base.repository.get_object_content("results.json")
        raw = json.loads(file)
        return get_data_from_raw(raw)
    return {}


def get_data_from_remote(source: RemoteData) -> dict:
    """Return fetched snapshot as a post-processed dictionary.

    Parameters
    ----------
    `source` : `RemoteData`
        The node of the remote folder containing the snapshot.

    Returns
    -------
    `dict`
        The post-processed data dictionary.
    """
    try:
        remote_path = source.get_attribute("remote_path")
        with open(f"{remote_path}/snapshot.json") as file:
            return get_data_from_file(file)
    except Exception:
        return {}


def add_analysis(data: dict) -> Styler | str:
    """Return analysis details.

    Parameters
    ----------
    `data` : `dict`
        The post-processed data dictionary.

    Returns
    -------
    `Styler | str`
        The details of the analysis.
    """

    if data:

        COLUMNS = {
            "time": "Time (s)",
            "I": "I (A)",
            "Ewe": "Ewe (V)",
        }

        selected_keys = {key: data[key] for key in COLUMNS}
        df = DataFrame(selected_keys).rename(columns=COLUMNS)

        return df.style.set_properties(width="100vw").set_table_styles([
            dict(
                selector="th, td",
                props=[
                    ("text-align", "center"),
                ],
            ),
        ]).hide(axis="index")

    else:
        return "ERROR! Failed to find or parse output"



"""pytest fixtures for simplified testing."""
import pytest

pytest_plugins = ["aiida.manage.tests.pytest_fixtures"]


@pytest.fixture(scope="function", autouse=True)
def clear_database_auto(clear_database):  # pylint: disable=unused-argument
    """Automatically clear database in between tests."""


@pytest.fixture(scope="function")
def aurora_code(aiida_local_code_factory):
    """Get a aurora code."""
    return aiida_local_code_factory(executable="diff", entry_point="aurora")



"""
STATUS: Providing the parameters as a dict with the specific parameter dict
as a default is not allowing me to initialize the CyclingTechnique's with
a normal dictionary (the parameters dict is not being correctly converted
to the right type but to Generic[DataT] I think).

Using TypedDict as suggested here is not working properly either:
https://stackoverflow.com/q/74643755/638366

I probably need to change something fundamentally with how the CyclingTechnique
and CyclingParameter interact with each other but I'm not sure what currently,
so for now I'm defining the InternalParameters as its own class inside each of
the CyclingTechnique's.

I also need to add items and __getitem__ because some other parts of the code
expect it to behave like a dict...
"""

from typing import Dict, Generic, Literal, Optional, Sequence, TypeVar, Union, get_args

from pydantic import BaseModel, Extra, NonNegativeFloat, NonNegativeInt
from pydantic.generics import GenericModel

DataT = TypeVar('DataT')


class CyclingParameter(GenericModel, Generic[DataT]):
    "Cycling parameter of type DataT"
    label: str  # the label used in a widget
    description: str = ""  # a long description
    units: str = ""  # physical units of this parameter
    value: Optional[DataT] = None  # the set value
    default_value: Optional[DataT] = None  # the default value
    required: bool = False  # True if parameter is required

    class Config:
        validate_assignment = True
        extra = Extra.forbid

    # @classmethod
    # def __concrete_name__(cls: Type[Any], params: Tuple[Type[Any], ...]) -> str:
    # return f'{params[0].__name__.title()}CyclingParameter'


class CyclingTechnique(BaseModel):
    device: Literal["worker", "MPG2"]  # the device name
    technique: str  # the technique name for tomato
    short_name: str  # short name for the technique
    name: str  # a custom name
    description: str = ""
    parameters: Dict[str, CyclingParameter]

    class Config:
        validate_assignment = True
        extra = Extra.forbid


allowed_I_ranges = Literal["keep",
                           "100 pA",
                           "1 nA",
                           "10 nA",
                           "100 nA",
                           "1 uA",
                           "10 uA",
                           "100 uA",
                           "1 mA",
                           "10 mA",
                           "100 mA",
                           "1 A",
                           "booster",
                           "auto",
                           ]
allowed_E_ranges = Literal["+-2.5 V", "+-5.0 V", "+-10 V", "auto",]

################################################################################
# SPECIFIC TECHNIQUES
################################################################################


class DummySequential(CyclingTechnique):
    device: Literal["worker"] = "worker"
    technique: Literal["sequential"] = "sequential"
    short_name: Literal["DUMMY_SEQUENTIAL"] = "DUMMY_SEQUENTIAL"
    name = "Dummy Sequential"
    description = "Dummy worker - sequential numbers"

    class InternalParameters(BaseModel):
        time: CyclingParameter[NonNegativeFloat]
        delay: CyclingParameter[NonNegativeFloat]

        @property
        def items(self):
            return dict(self._iter()).items

        def __getitem__(self, item):
            return getattr(self, item)

    # parameters: Dict[str, CyclingParameter] = {
    parameters: InternalParameters = InternalParameters(
        **{
            "time": CyclingParameter[NonNegativeFloat](label="Time:", units="s", default_value=100., required=True),
            "delay": CyclingParameter[NonNegativeFloat](
                label="Delay:",
                units="s",
                default_value=1.0,
                required=True,
            )
        }
    )


################################################################################
class DummyRandom(CyclingTechnique):
    device: Literal["worker"] = "worker"
    technique: Literal["random"] = "random"
    short_name: Literal["DUMMY_RANDOM"] = "DUMMY_RANDOM"
    name = "Dummy Random"
    description = "Dummy worker - random numbers"

    class InternalParameters(BaseModel):
        time: CyclingParameter[NonNegativeFloat]
        delay: CyclingParameter[NonNegativeFloat]

        @property
        def items(self):
            return dict(self._iter()).items

        def __getitem__(self, item):
            return getattr(self, item)

    # parameters: Dict[str, CyclingParameter] = {
    parameters: InternalParameters = InternalParameters(
        **{
            "time": CyclingParameter[NonNegativeFloat](label="Time:", units="s", default_value=100., required=True),
            "delay": CyclingParameter[NonNegativeFloat](
                label="Delay:",
                units="s",
                default_value=1.0,
                required=True,
            )
        }
    )


################################################################################
class OpenCircuitVoltage(CyclingTechnique):
    device: Literal["MPG2"] = "MPG2"
    technique: Literal["open_circuit_voltage"] = "open_circuit_voltage"
    short_name: Literal["OCV"] = "OCV"
    name = "OCV"
    description = "Open circuit voltage"

    class InternalParameters(BaseModel):
        time: CyclingParameter[NonNegativeFloat]
        record_every_dt: CyclingParameter[NonNegativeFloat]
        record_every_dE: CyclingParameter[NonNegativeFloat]
        I_range: CyclingParameter[allowed_I_ranges]
        E_range: CyclingParameter[allowed_E_ranges]

        @property
        def items(self):
            return dict(self._iter()).items

        def __getitem__(self, item):
            return getattr(self, item)

    # parameters: Dict[str, CyclingParameter] = {
    parameters: InternalParameters = InternalParameters(
        **{
            "time":
            CyclingParameter[NonNegativeFloat](
                label="Time:",
                description="The length of the OCV step",
                units="s",
                default_value=0.0,
                required=True,
            ),
            "record_every_dt":
            CyclingParameter[NonNegativeFloat](
                label="Record every $dt$:",
                description="Record a datapoint at prescribed time spacing",
                units="s",
                default_value=30.0,
                required=True,
            ),
            "record_every_dE":
            CyclingParameter[NonNegativeFloat](
                label="Record every $dE$:",
                description="Record a datapoint at prescribed voltage spacing",
                units="V",
                default_value=0.005,
                required=True,
            ),
            "I_range":
            CyclingParameter[allowed_I_ranges](
                label="I range",
                description="",
                # TODO: "keep" value does not work - setting to 1 A
                default_value="1 A",
                required=True,
            ),
            "E_range":
            CyclingParameter[allowed_E_ranges](
                label="E range",
                description="",
                default_value="auto",
                required=True,
            ),
        }
    )


################################################################################
class ConstantVoltage(CyclingTechnique):
    device: Literal["MPG2"] = "MPG2"
    technique: Literal["constant_voltage"] = "constant_voltage"
    short_name: Literal["CV"] = "CV"
    name = "CV"
    description = "Controlled voltage technique, with optional current and voltage limits"

    class InternalParameters(BaseModel):
        time: CyclingParameter[NonNegativeFloat]
        voltage: CyclingParameter[float]
        record_every_dt: CyclingParameter[NonNegativeFloat]
        record_every_dI: CyclingParameter[NonNegativeFloat]
        I_range: CyclingParameter[allowed_I_ranges]
        E_range: CyclingParameter[allowed_E_ranges]
        n_cycles: CyclingParameter[NonNegativeInt]
        is_delta: CyclingParameter[bool]
        exit_on_limit: CyclingParameter[bool]
        limit_voltage_max: CyclingParameter[float]
        limit_voltage_min: CyclingParameter[float]
        limit_current_max: CyclingParameter[Union[float, str]]
        limit_current_min: CyclingParameter[Union[float, str]]

        @property
        def items(self):
            return dict(self._iter()).items

        def __getitem__(self, item):
            return getattr(self, item)

    # parameters: Dict[str, CyclingParameter] = {
    parameters: InternalParameters = InternalParameters(
        **{
            "time":
            CyclingParameter[NonNegativeFloat](
                label="Time:",
                description="Maximum duration of the CV step",
                units="s",
                default_value=0.0,
                required=True,
            ),
            "voltage":
            CyclingParameter[float](
                label="Step voltage:",
                description="Voltage of the current step",
                units="V",
                default_value=0.0,
                required=True,
            ),
            "record_every_dt":
            CyclingParameter[NonNegativeFloat](
                label="Record every $dt$:",
                description="Record a datapoint at prescribed time spacing",
                units="s",
                default_value=30.0,
                required=True,
            ),
            "record_every_dI":
            CyclingParameter[NonNegativeFloat](
                label="Record every $dI$:",
                description="Record a datapoint at prescribed current spacing",
                units="I",
                default_value=0.001,
                required=True,
            ),
            "I_range":
            CyclingParameter[allowed_I_ranges](
                label="I range",
                description="Select the current range of the potentiostat",
                default_value="keep",
                required=True,
            ),
            "E_range":
            CyclingParameter[allowed_E_ranges](
                label="E range",
                description="Select the voltage range of the potentiostat",
                default_value="auto",
                required=True,
            ),
            "n_cycles":
            CyclingParameter[NonNegativeInt](
                label="Number of cycles:",
                description="Cycle through the current technique N times.",
                default_value=0,
                required=True,
            ),
            "is_delta":
            CyclingParameter[bool](
                label="Δ$V$:",
                description=r"""
                    Is the step voltage a $\Delta$ from previous step?
                """,
                default_value=False,
                required=True,
            ),
            "exit_on_limit":
            CyclingParameter[bool](
                label="Exit when limits reached?",
                description="Stop the whole experiment when limit is reached?",
                default_value=False,
                required=True,
            ),
            "limit_voltage_max":
            CyclingParameter[float](
                label="Maximum voltage:",
                description="Define the upper limit of voltage for this step",
                units="V",
                default_value=None
            ),
            "limit_voltage_min":
            CyclingParameter[float](
                label="Minimum voltage:",
                description="Define the lower limit of voltage for this step",
                units="V",
                default_value=None
            ),
            "limit_current_max":
            CyclingParameter[Union[float, str]](
                label="Maximum current:",
                description="Define the upper limit of current for this step",
                units="I",
                default_value=None
            ),
            "limit_current_min":
            CyclingParameter[Union[float, str]](
                label="Minimum current:",
                description="Define the lower limit of current for this step",
                units="I",
                default_value=None
            )
        }
    )


################################################################################
class ConstantCurrent(CyclingTechnique):
    device: Literal["MPG2"] = "MPG2"
    technique: Literal["constant_current"] = "constant_current"
    short_name: Literal["CC"] = "CC"
    name = "CC"
    description = "Controlled current technique, with optional voltage and current limits"

    class InternalParameters(BaseModel):
        time: CyclingParameter[NonNegativeFloat]
        current: CyclingParameter[Union[float, str]]
        record_every_dt: CyclingParameter[NonNegativeFloat]
        record_every_dE: CyclingParameter[NonNegativeFloat]
        I_range: CyclingParameter[allowed_I_ranges]
        E_range: CyclingParameter[allowed_E_ranges]
        n_cycles: CyclingParameter[NonNegativeInt]
        is_delta: CyclingParameter[bool]
        exit_on_limit: CyclingParameter[bool]
        limit_voltage_max: CyclingParameter[float]
        limit_voltage_min: CyclingParameter[float]
        limit_current_max: CyclingParameter[Union[float, str]]
        limit_current_min: CyclingParameter[Union[float, str]]

        @property
        def items(self):
            return dict(self._iter()).items

        def __getitem__(self, item):
            return getattr(self, item)

    # parameters: Dict[str, CyclingParameter] = {
    parameters: InternalParameters = InternalParameters(
        **{
            "time":
            CyclingParameter[NonNegativeFloat](
                label="Time:",
                description="Maximum duration of the CC step",
                units="s",
                default_value=0.0,
                required=True,
            ),
            "current":
            CyclingParameter[Union[float, str]](
                label="Step current:",
                description="Current during the current step",
                units="I",
                default_value=0.0,
                required=True,
            ),
            "record_every_dt":
            CyclingParameter[NonNegativeFloat](
                label="Record every $dt$:",
                description="Record a datapoint at prescribed time spacing",
                units="s",
                default_value=30.0,
                required=True,
            ),
            "record_every_dE":
            CyclingParameter[NonNegativeFloat](
                label="Record every $dE$:",
                description="Record a datapoint at prescribed voltage spacing",
                units="V",
                default_value=0.005,
                required=True,
            ),
            "I_range":
            CyclingParameter[allowed_I_ranges](
                label="I range",
                description="Select the current range of the potentiostat",
                default_value="keep",
                required=True,
            ),
            "E_range":
            CyclingParameter[allowed_E_ranges](
                label="E range",
                description="Select the voltage range of the potentiostat",
                default_value="auto",
                required=True,
            ),
            "n_cycles":
            CyclingParameter[NonNegativeInt](
                label="Number of cycles:",
                description="Cycle through the current technique N times.",
                default_value=0,
                required=True,
            ),
            "is_delta":
            CyclingParameter[bool](
                label="Δ$I$:",
                description=r"""
                    Is the step current a $\Delta$ from previous step?
                """,
                default_value=False,
                required=True,
            ),
            "exit_on_limit":
            CyclingParameter[bool](
                label="Exit when limits reached?",
                description="Stop the whole experiment when limit is reached?",
                default_value=False,
                required=True,
            ),
            "limit_voltage_max":
            CyclingParameter[float](
                label="Maximum voltage:",
                description="Define the upper limit of voltage for this step",
                units="V",
                default_value=None
            ),
            "limit_voltage_min":
            CyclingParameter[float](
                label="Minimum voltage:",
                description="Define the lower limit of voltage for this step",
                units="V",
                default_value=None
            ),
            "limit_current_max":
            CyclingParameter[Union[float, str]](
                label="Maximum current:",
                description="Define the upper limit of current for this step",
                units="I",
                default_value=None
            ),
            "limit_current_min":
            CyclingParameter[Union[float, str]](
                label="Minimum current:",
                description="Define the lower limit of current for this step",
                units="I",
                default_value=None
            )
        }
    )


################################################################################
# class SweepVoltage(CyclingTechnique):
#    technique: Literal["sweep_voltage"] = "sweep_voltage"
#    short_name: Literal["LSV"] = "LSV"
#    name = "LSV"
#    description = "Controlled voltage technique, allowing linear change of voltage between pre-defined endpoints as a function of time, with optional current and voltage limits"

################################################################################
# class SweepCurrent(CyclingTechnique):
#    technique: Literal["sweep_current"] = "sweep_current"
#    short_name: Literal["LSC"] = "LSC"
#    name = ""
#    description = "Controlled current technique, allowing linear change of current between pre-defined endpoints as a function of time, with optional current and voltage limits"


################################################################################
class Loop(CyclingTechnique):
    device: Literal["MPG2"] = "MPG2"
    technique: Literal["loop"] = "loop"
    short_name: Literal["LOOP"] = "LOOP"
    name = "LOOP"
    description = "Loop technique, allowing to repeat a set of preceding techniques in a technique array"

    class InternalParameters(BaseModel):
        n_gotos: CyclingParameter[int]
        goto: CyclingParameter[int]

        @property
        def items(self):
            return dict(self._iter()).items

        def __getitem__(self, item):
            return getattr(self, item)

    # parameters: Dict[str, CyclingParameter] = {
    parameters: InternalParameters = InternalParameters(
        **{
            "n_gotos":
            CyclingParameter[int](
                label="Repeats",
                description="""
                    Number of times the technique will jump; set to -1 for unlimited
                """,
                default_value=-1,
                required=True,
            ),
            "goto":
            CyclingParameter[int](
                label="Go to:",
                description="Index of the technique to go back to",
                default_value=1,
                required=True,
            ),
        }
    )


################################################################################
# SEQUENCE OF TECHNIQUES
################################################################################

ElectroChemPayloads = Union[DummySequential,
                            DummyRandom,
                            OpenCircuitVoltage,
                            ConstantCurrent,
                            ConstantVoltage,
                            #    SweepCurrent,
                            #    SweepVoltage,
                            Loop,
                            ]


class ElectroChemSequence(BaseModel):
    name: str = ""
    method: Sequence[ElectroChemPayloads]

    class Config:
        validate_assignment = True
        extra = Extra.forbid

    @property
    def n_steps(self):
        "Number of steps of the method"
        return len(self.method)

    def set_name(self, name: str) -> None:
        """docstring"""
        self.name = name

    def add_step(self, elem):
        if not isinstance(elem, get_args(ElectroChemPayloads)):
            raise ValueError("Invalid technique")
        self.method.append(elem)

    def remove_step(self, index):
        self.method.pop(index)

    def move_step_backward(self, i):
        if i > 0:
            j = i - 1
            self.method[j], self.method[i] = self.method[i], self.method[j]

    def move_step_forward(self, i):
        if i < self.n_steps - 1:
            j = i + 1
            self.method[i], self.method[j] = self.method[j], self.method[i]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ElectroChemSequence):
            return NotImplemented
        return self.name == other.name



from json import load
from tempfile import NamedTemporaryFile
from typing import Optional

from aiida.orm import CalcJobNode
from aiida.transports import Transport

from .utils.analyzers import CapacityAnalyzer


def monitor_capacity_threshold(
    node: CalcJobNode,
    transport: Transport,
    settings: dict,
    filename="snapshot.json",
) -> Optional[str]:
    """Retrieve and inspect snapshot to determine if capacity has
    fallen below threshold for several consecutive cycles.

    Parameters
    ----------
    `node` : `CalcJobNode`
        The calculation node.
    `transport` : `Transport`
        The associated transport instance.
    `settings` : `dict`
        The monitor settings.
    `filename` : `str`
        The polled source file, `"snapshot.json"` by default.

    Returns
    -------
    `Optional[str]`
        If condition is met, an exit message, `None` otherwise.

    Raises
    ------
    `TypeError`
        If source file is not in expected dictionary format (JSON).
    `ValueError`
        If source file is empty.
    `FileNotFoundError`
        If the file does not exist in the working directory.
    `OSError`
        If another error occurred while reading the file.
    `Exception`
        If something else prevented analysis.
    """

    analyzer = CapacityAnalyzer(**settings)
    analyzer.set_logger(node.logger)

    try:

        with transport:

            remote_path = f"{node.get_remote_workdir()}/{filename}"

            if not transport.isfile(remote_path):
                node.logger.info(f"'{filename}' not yet produced; continue")
                return None

            try:

                with NamedTemporaryFile("w+") as temp_file:
                    transport.getfile(remote_path, temp_file.name)
                    snapshot = load(temp_file)

                if not isinstance(snapshot, dict):
                    raise TypeError

                if not snapshot:
                    raise ValueError

                return analyzer.analyze(snapshot)

            except TypeError:
                node.logger.error(f"'{filename}' not in dictionary format")
            except ValueError:
                node.logger.error(f"'{filename}' is empty")
            except FileNotFoundError:
                node.logger.error(f"error fetching '{filename}'")
            except OSError as err:
                node.logger.error(str(err))

            return None

    except Exception as err:
        node.logger.error(str(err))
        return None



from pydantic import BaseModel

from aiida_aurora.schemas.cycling import ElectroChemSequence


def electrochemsequence_to_method_list_0(elchemsequence: ElectroChemSequence, MethodSchema: BaseModel):
    """
    Convert an ElectroChemSequence into a list of Method.
    Parameter values that are None will be ignored.

    Compatible with the following dgbowl-schemas payload versions:
        [0.1, 0.2]
    """

    if not isinstance(elchemsequence, ElectroChemSequence):
        if isinstance(elchemsequence, dict):
            elchemsequence = ElectroChemSequence(**elchemsequence)
        else:
            raise TypeError()
    sequence = []
    for step in elchemsequence.method:
        parameters = {name: param.value for name, param in step.parameters.items() if param.value is not None}

        sequence.append(MethodSchema(**{
            'device': step.device,
            'technique': step.technique,
        }, **parameters))

    return sequence



#!/usr/bin/env python
"""This script emulates the Aurora's server API responding to requests."""

import argparse
import datetime
import json
from json import JSONDecodeError

import numpy as np


def measure(time, record_every_dt):
    """Simulates the measure of V and I."""
    times = np.arange(0.0, time + 0.0001, record_every_dt)
    data = {
        "time": times.tolist(),
        "V": np.sin(times).tolist(),
        "I": np.cos(times).tolist(),
    }
    return data


def execute_experiment(battery_d, experiment_d):
    """Execute the experiment and return a dictionary."""
    start_time = datetime.datetime.now()
    print(f"I am executing an experiment at {start_time}. My inputs are:")
    print(f"battery_d    = {battery_d}")
    print(f"experiment_d = {experiment_d}")

    # validate inputs
    if experiment_d["description"].get("technique") not in ("OCV", "CPLIMIT"):
        raise ValueError("Technique not valid.")

    # perform measurement
    output_data = measure(
        experiment_d["parameters"]["time"],
        experiment_d["parameters"]["record_every_dt"],
    )
    end_time = start_time + datetime.timedelta(seconds=output_data["time"][-1] + 1)

    # generate output dic
    results_d = {
        "start_time":
        str(start_time),
        "sequence": [
            {
                "technique": experiment_d["description"].get("technique"),
                "parameters": experiment_d["parameters"],
                "duration": output_data["time"][-1],
                "output_data": output_data,
            },
        ],
        "end_time":
        str(end_time),
    }
    return results_d


def main():
    """Aurora API emulator."""
    parser = argparse.ArgumentParser(
        description="This script emulates the Aurora's server API responding to requests.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("battery_specs", type=str, help="A JSON file with the battery specifications.")
    parser.add_argument(
        "experiment_specs",
        type=str,
        help="A JSON file with the experiment specifications.",
    )
    parser.add_argument("-o", "--output", type=str, default="exp.out", help="Name of the output file.")
    args = parser.parse_args()

    try:
        battery_specs = json.load(open(args.battery_specs))
        experiment_specs = json.load(open(args.experiment_specs))
    except JSONDecodeError as err:
        raise RuntimeError("Failed to open JSON file.") from err

    output_d = {}
    # try:
    output_d = execute_experiment(battery_specs, experiment_specs)
    # except Exception as err:
    #    print(f'Error during experiment: {type(err)}: {err}')
    #    output_d['exit_status'] = 1
    #    output_d['exit_msg'] = err
    # else:
    #    output_d['exit_status'] = 0
    #    output_d['exit_msg'] = None

    print(f"I am writing results to {args.output} ...")
    json.dump(output_d, open(args.output, "w"))
    print("Done.")


if __name__ == "__main__":
    main()



""" Tests for calculations

"""
from aiida.plugins import CalculationFactory, DataFactory


def test_entry_points():
    """Tests the loading of the entry points."""
    DataFactory("aurora.batterysample")
    DataFactory("aurora.batterystate")
    DataFactory("aurora.cyclingspecs")
    DataFactory("aurora.tomatosettings")

    CalculationFactory("aurora.fake")
    CalculationFactory("aurora.cycler")



from datetime import datetime
from enum import Flag
from typing import Literal, Optional

from pydantic import BaseModel, NonNegativeFloat, PositiveFloat, root_validator

from .utils import extract_schema_types


class BatteryComposition(BaseModel):  # TypedDict?
    description: Optional[str] = None
    cathode: Optional[str] = None
    anode: Optional[str] = None
    electrolyte: Optional[str] = None

    class Config:
        # exclude fields from export, to avoid validation errors when reloaded
        fields = {
            'cathode': {
                'exclude': True
            },
            'anode': {
                'exclude': True
            },
            'electrolyte': {
                'exclude': True
            },
        }

    @root_validator
    def validate_composition(cls, values):
        """
        Check that components are not specified if 'description' is specified
        then build components from a 'description' string, or vice versa.
        # TODO: what to do if 'description' is not in the C|E|A format?
        """
        if values['description']:
            if any(values[key] for key in ('cathode', 'anode', 'electrolyte')):
                raise ValueError("You cannot specify a 'description' and any component at the same time.")
            values['description'] = values['description'].strip()
            components = list(map(str.strip, values['description'].split('|')))
            if len(components) == 3:
                values['cathode'], values['electrolyte'], values['anode'] = components
            else:
                values['cathode'], values['electrolyte'], values['anode'] = (None, None, None)
                # raise ValueError(
                # "Composition 'description' does not have 3 components (i.e. {cathode}|{electrolyte}|{anode}).")
        elif any(values[key] for key in ('cathode', 'anode', 'electrolyte')):
            for key in ('cathode', 'anode', 'electrolyte'):
                values[key] = values[key].strip()
            values['description'] = f"{values['cathode']}|{values['electrolyte']}|{values['anode']}"
        else:
            raise ValueError("You must specify either a string 'description' or the components.")
        return values


class BatteryCapacity(BaseModel):  # TypedDict?
    nominal: PositiveFloat
    actual: Optional[NonNegativeFloat]
    units: Literal["mAh", "Ah"]


class BatteryMetadata(BaseModel):
    name: str
    creation_datetime: datetime
    creation_process: str


class BatterySpecs(BaseModel):
    """
    Battery specification schema.
    """
    manufacturer: str
    composition: BatteryComposition
    form_factor: str
    capacity: BatteryCapacity

    # manufacturer, form_factor:
    # should we use a Literal or a validator to check that they are one of the available ones?

    # add pre-validator to specify capacity as str (e.g. '4.8 mAh')?


class BatterySample(BatterySpecs):
    """
    Battery sample schema.
    """
    battery_id: int
    metadata: BatteryMetadata


class ChargeState(Flag):
    """Defines the charge state of a battery."""
    CHARGED = True
    DISCHARGED = False


class BatteryState(BatterySample):
    """
    Battery state schema.
    """
    state_id: int
    charged: ChargeState = ChargeState.CHARGED


BatterySpecsJsonTypes = extract_schema_types(BatterySpecs)
BatterySampleJsonTypes = extract_schema_types(BatterySample)



"""
Command line interface (cli) for aiida_aurora.

Register new commands either via the "console_scripts" entry point or plug them
directly into the 'verdi' command by using AiiDA-specific entry points like
"aiida.cmdline.data" (both in the setup.json file).
"""

import sys

import click

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params.types import DataParamType
from aiida.cmdline.utils import decorators
from aiida.orm import QueryBuilder
from aiida.plugins import DataFactory


# See aiida.cmdline.data entry point in setup.json
@verdi_data.group("aurora")
def data_cli():
    """Command line interface for aiida-aurora"""


@data_cli.command("list")
@decorators.with_dbenv()
def list_():  # pylint: disable=redefined-builtin
    """
    Display all DiffParameters nodes
    """
    DiffParameters = DataFactory("aurora")

    qb = QueryBuilder()
    qb.append(DiffParameters)
    results = qb.all()

    s = ""
    for result in results:
        obj = result[0]
        s += f"{str(obj)}, pk: {obj.pk}\n"
    sys.stdout.write(s)


@data_cli.command("export")
@click.argument("node", metavar="IDENTIFIER", type=DataParamType())
@click.option(
    "--outfile",
    "-o",
    type=click.Path(dir_okay=False),
    help="Write output to file (default: print to stdout).",
)
@decorators.with_dbenv()
def export(node, outfile):
    """Export a DiffParameters node (identified by PK, UUID or label) to plain text."""
    string = str(node)

    if outfile:
        with open(outfile, "w") as f:
            f.write(string)
    else:
        click.echo(string)



from pydantic import BaseModel

from aiida_aurora.schemas.battery import BatterySample


def batterysample_to_sample_0(batsample: BatterySample, SampleSchema: BaseModel):
    """
    Convert a BatterySample into a Sample.

    Compatible with the following dgbowl-schemas payload versions:
        [0.1, 0.2]
    """

    if not isinstance(batsample, BatterySample):
        if isinstance(batsample, dict):
            batsample = BatterySample(**batsample)
        else:
            raise TypeError()
    # if batsample.capacity.units == "mAh":
    # capacity = float(batsample.capacity.nominal) * 0.001
    # elif batsample.capacity.units == "Ah":
    # capacity = float(batsample.capacity.nominal)

    sample = SampleSchema(name=batsample.metadata.name, capacity=batsample.capacity.nominal)
    return sample



import functools

from pandas import DataFrame, Series, isna

from .. import battery as battery_schemas


def _remove_empties_from_dict(a_dict):
    # this may not work for nested lists
    new_dict = {}
    for k, v in a_dict.items():
        if isinstance(v, dict):
            v = _remove_empties_from_dict(v)
        elif isinstance(v, list):  # only works for lists of dict
            new_dict[k] = [_remove_empties_from_dict(vv) for vv in v]
            continue
        if v is not None and not isna(v) and v != "":
            new_dict[k] = v
    return new_dict


def remove_empties_from_dict_decorator(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        dic = func(*args, **kwargs)
        if isinstance(dic, list):
            return [_remove_empties_from_dict(d) for d in dic]
        return _remove_empties_from_dict(dic)

    return wrapper


def _make_formatted_dict(my_dict, key_arr, val):
    """
    Set val at path in my_dict defined by the string (or serializable object) array key_arr
    """
    current = my_dict
    for i in range(len(key_arr)):
        key = key_arr[i]
        if key not in current:
            if i == len(key_arr) - 1:
                current[key] = val
            else:
                current[key] = {}
        else:
            if type(current[key]) is not dict:
                print("Given dictionary is not compatible with key structure requested")
                raise ValueError("Dictionary key already occupied")
        current = current[key]
    # return _remove_empties_from_dict(my_dict)
    return my_dict


def pd_dataframe_to_formatted_json(df, sep="."):
    """Convert a pandas.DataFrame to a list of nested dictionaries."""
    if not isinstance(df, DataFrame):
        raise TypeError('df should be a pandas.DataFrame object')
    result = []
    for _, row in df.iterrows():
        parsed_row = {}
        for idx, val in row.iteritems():
            keys = idx.split(sep)
            parsed_row = _make_formatted_dict(parsed_row, keys, val)
        result.append(parsed_row)
    return result


def dict_to_formatted_json(series, sep="."):
    """Convert a flat dictionary or a pandas.Series to a nested dictionary."""
    if not isinstance(series, (dict, Series)):
        raise TypeError('series should be a pandas.Series object')
    parsed_series = {}
    for idx, val in series.items():
        keys = idx.split(sep)
        parsed_series = _make_formatted_dict(parsed_series, keys, val)
    return parsed_series


def extract_schema_types(model, sep="."):
    """Convert a pydantic schema into a nested dictionary containing types."""
    SCHEMA_TYPES = {'string': str, 'integer': int, 'number': float}
    schema_dic = {}
    for name, sdic in model.schema()['properties'].items():
        if '$ref' in sdic:
            # print(f"  {name} is a class {sdic['$ref']}")
            sub_schema = extract_schema_types(
                getattr(battery_schemas, sdic['$ref'].split('/')[-1])
            )  # call extract_schema for the $ref class
            for key, value in sub_schema.items():
                schema_dic[f'{name}.{key}'] = value
        elif 'type' in sdic:
            # print(f"{name} is of type {sdic['type']}")
            schema_dic[name] = SCHEMA_TYPES[sdic['type']]
    return schema_dic



"""
Parsers provided by aiida_aurora.

Register parsers via the "aiida.parsers" entry point in setup.json.
"""
import json
import os
import re

import numpy as np

from aiida.common import exceptions
from aiida.engine import ExitCode
from aiida.orm import ArrayData, SinglefileData
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory

BatteryCyclerExperiment = CalculationFactory("aurora.cycler")


class TomatoParser(Parser):
    """
    Parser class for parsing the output of a BatteryCyclerExperiment generated by tomato.

    Two files should be found in the working directory:

      - a ZIP file, containing the raw data

      - a JSON file, containing the data pre-processed by tomato & yadg

    Priority is to retrieve the parsed JSON file.
    The raw data ZIP file is also retained if found.
    """

    def __init__(self, node):
        """
        Initialize Parser instance

        Checks that the ProcessNode being passed was produced by a BatteryCyclerExperiment.

        :param node: ProcessNode of calculation
        :param type node: :class:`aiida.orm.ProcessNode`
        """
        super().__init__(node)
        if not issubclass(node.process_class, BatteryCyclerExperiment):
            raise exceptions.ParsingError("Can only parse BatteryCyclerExperiment")

    def parse(self, **kwargs):
        """
        Parse outputs, store results in database.

        :returns: an exit code, if parsing fails (or nothing if parsing succeeds)
        """
        retrieved_temporary_folder = kwargs["retrieved_temporary_folder"]
        output_json_filename = self.node.get_option("output_filename") + ".json"
        output_zip_filename = os.path.join(retrieved_temporary_folder, self.node.get_option("output_filename") + ".zip")

        files_retrieved = self.retrieved.list_object_names()

        # Check that zip file is present
        if os.path.isfile(output_zip_filename):
            try:
                self.logger.debug(f"Storing '{output_zip_filename}'")
                output_raw_data_node = SinglefileData(output_zip_filename)
                self.out("raw_data", output_raw_data_node)
                output_raw_data_node_created = True
            except Exception:
                self.logger.warning(f"The raw data zip file '{output_zip_filename}' could not be read.")
                output_raw_data_node_created = False
        else:
            self.logger.warning(f"The raw data zip file '{output_zip_filename}' is missing.")
            output_raw_data_node_created = False

        # Check that json file is present
        if output_json_filename not in files_retrieved:
            self.logger.error(f"The output json file '{output_json_filename}' is missing.")
            if output_raw_data_node_created:
                # only json file is missing
                return self.exit_codes.ERROR_OUTPUT_JSON_MISSING
            # both files are missing
            return self.exit_codes.ERROR_OUTPUT_FILES_MISSING

        # If a json file was found, parse it and add output node
        try:
            self.logger.debug(f"Parsing '{output_json_filename}'")
            with self.retrieved.open(output_json_filename, "r") as handle:
                output_results_node = self.parse_tomato_results(json.load(handle), self.logger)
            self.out("results", output_results_node)
        except OSError:
            self.logger.error(f"Error opening the json file '{output_json_filename}'.")
            return self.exit_codes.ERROR_OUTPUT_JSON_READ
        except json.JSONDecodeError:
            self.logger.error(f"Error parsing json file '{output_json_filename}'.")
            return self.exit_codes.ERROR_OUTPUT_JSON_PARSE

        # If found, the zip file was already stored in a SinglefileData node
        if not output_raw_data_node_created:
            return self.exit_codes.WARNING_OUTPUT_ZIP_MISSING

        # If files were retrieved and parsed by the job was completed with error or cancelled,
        # use the specific error codes
        job_annotation = self.node.get_attribute("last_job_info").get("annotation")
        if job_annotation == "Completed with error":
            return self.exit_codes.WARNING_COMPLETED_ERROR
        elif job_annotation == "Cancelled":
            return self.exit_codes.WARNING_COMPLETED_CANCELLED

        return ExitCode(0)

    @staticmethod
    def parse_tomato_results(data_dic, logger=None):
        """
        Parse results.json file.

        :returns: a :class:`aiida.orm.ArrayData` in this way:

          - `metadata` is stored as attribute

          - `data` is split in steps, physical quantity name, and n/s/u identifier (nominal value, std error, units)
            The name of each array is:  `'step{step_number}_{raw_quantity_name}_{identifier}'`

        """

        parsed = {}

        data = data_dic["steps"][0]["data"]

        keys = ["Ewe", "I"]  # HACK hardcoded
        fill = {"n": np.nan, "s": np.nan, "u": ""}

        if logger:
            logger.debug(f"parse_tomato_results: storing {keys}")

        for key in keys:
            clean_key = re.sub("[^0-9a-zA-Z_]", "_", key)  # TODO necessary?
            for id in ("n", "s", "u"):
                values = [step["raw"].get(key, fill)[id] for step in data]
                parsed[f"step0_{clean_key}_{id}"] = np.array(values)

        parsed["step0_uts"] = np.array([step["uts"] for step in data])

        node = ArrayData()
        for key, value in parsed.items():
            node.set_array(key, value)
            node.set_attribute_many(data_dic["metadata"])

        if logger:
            logger.debug(f"parse_tomato_results: {list(parsed.keys())} stored")

        return node



"""
Dummy CalcJobs to simulate a battery experiment.
"""
from aiida.common import datastructures
from aiida.engine import CalcJob
# from aiida.orm import SinglefileData
from aiida.orm import Dict

from aiida_aurora.data.battery import BatterySampleData
# TODO: from aiida_aurora.data.battery import BatteryStateData
from aiida_aurora.data.experiment import CyclingSpecsData


class BatteryFakeExperiment(CalcJob):
    """
    AiiDA calculation plugin for the fake_aurora_server script.

    Simple AiiDA plugin that sends input data as json files to the fake Aurora server API script.
    """

    _INPUT_BATTERY_JSON_FILE = "battery.json"
    _INPUT_EXPERIMENT_JSON_FILE = "experiment.json"
    _OUTPUT_JSON_FILE = "output.json"
    _DEFAULT_STDOUT_FILE = "output.log"

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs["metadata"]["options"]["resources"].default = {
            "num_machines": 1,
            "num_mpiprocs_per_machine": 1,
        }
        spec.inputs["metadata"]["options"]["parser_name"].default = "aurora"

        # new ports
        spec.input(
            "metadata.options.output_filename",
            valid_type=str,
            default=cls._DEFAULT_STDOUT_FILE,
        )
        spec.input("battery_sample", valid_type=BatterySampleData, help="Battery sample used.")
        spec.input("exp_specs", valid_type=CyclingSpecsData, help="Experiment specifications.")
        spec.output("results", valid_type=Dict, help="Results of the experiment.")  # a proper type should be defined
        # spec.output('battery_state', valid_type=BatteryStateData, help='State of the battery after the experiment.')

        spec.exit_code(
            300, "ERROR_MISSING_OUTPUT_FILES", message="Calculation did not produce all expected output files."
        )

    def prepare_for_submission(self, folder):
        """
        Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files
            needed by the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """

        with open(self._INPUT_BATTERY_JSON_FILE, "w") as handle:
            handle.write(self.inputs.battery_sample.get_json())
        with open(self._INPUT_EXPERIMENT_JSON_FILE, "w") as handle:
            handle.write(self.inputs.exp_specs.get_json())

        codeinfo = datastructures.CodeInfo()
        # codeinfo.cmdline_params = self.inputs.parameters.cmdline_params(
        #     file1_name=self.inputs.file1.filename,
        #     file2_name=self.inputs.file2.filename)
        codeinfo.cmdline_params = [
            self._INPUT_BATTERY_JSON_FILE,
            self._INPUT_EXPERIMENT_JSON_FILE,
            "-o",
            self._OUTPUT_JSON_FILE,
        ]
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.metadata.options.output_filename
        codeinfo.withmpi = self.inputs.metadata.options.withmpi

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.local_copy_list = []
        calcinfo.retrieve_list = [
            self.metadata.options.output_filename,
            self._OUTPUT_JSON_FILE,
        ]

        return calcinfo



"""
Calculations provided by aiida_aurora.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""

from .cycling_sequence import CyclingSequenceWorkChain

__all__ = [
    'CyclingSequenceWorkChain',
]



"""
Battery cycling experiment CalcJobs.
"""
from typing import Optional

import yaml

from aiida.common import datastructures
from aiida.engine import CalcJob
from aiida.engine.processes.exit_code import ExitCode
from aiida.orm import ArrayData, Node, SinglefileData

from aiida_aurora.data.battery import BatterySampleData
# TODO: from aiida_aurora.data.battery import BatteryStateData
from aiida_aurora.data.control import TomatoSettingsData
from aiida_aurora.data.experiment import CyclingSpecsData
from aiida_aurora.schemas.dgbowl import conversion_map, payload_models


class BatteryCyclerExperiment(CalcJob):
    """
    AiiDA calculation plugin for the tomato instrument automation package.
    https://github.com/dgbowl/tomato
    """

    _INPUT_PAYLOAD_YAML_FILE = "payload.yaml"
    _INPUT_PAYLOAD_VERSION = "0.2"
    _OUTPUT_FILE_PREFIX = "results"

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        super().define(spec)

        # set default values for AiiDA options
        spec.inputs["metadata"]["options"]["resources"].default = {  # REQUIRED?
            "num_machines": 1,
        }
        spec.inputs["metadata"]["options"]["parser_name"].default = "aurora"
        spec.inputs["metadata"]["options"]["withmpi"].default = False
        spec.inputs["metadata"]["options"]["input_filename"].default = cls._INPUT_PAYLOAD_YAML_FILE
        # spec.inputs['metadata']['options']['scheduler_stderr'].default = ''
        # spec.inputs['metadata']['options']['scheduler_stdout'].default = ''

        # new ports
        spec.input(
            "metadata.options.output_filename",
            valid_type=str,
            default=cls._OUTPUT_FILE_PREFIX,
        )
        spec.input("battery_sample", valid_type=BatterySampleData, help="Battery sample used.")
        spec.input("protocol", valid_type=CyclingSpecsData, help="Experiment specifications.")
        spec.input("control_settings", valid_type=TomatoSettingsData, help="Experiment control settings.")
        spec.output("results", valid_type=ArrayData, help="Results of the experiment.")
        spec.output("raw_data", valid_type=SinglefileData, help="Raw data retrieved.")
        # spec.output('battery_state', valid_type=BatteryStateData, help='State of the battery after the experiment.')

        # Unrecoverable errors: required retrieved files could not be read, parsed or are otherwise incomplete
        spec.exit_code(300, "ERROR_OUTPUT_FILES_MISSING", message="Experiment did not produce any kind of output file.")
        spec.exit_code(301, "ERROR_OUTPUT_JSON_MISSING", message="Experiment did not produce an output JSON file.")
        spec.exit_code(
            311, "ERROR_OUTPUT_JSON_READ", message="The output JSON file could not be read. Raw data may be available."
        )
        spec.exit_code(
            312,
            "ERROR_OUTPUT_JSON_PARSE",
            message="The output JSON file could not be parsed. Raw data may be available."
        )

        # Warnings: JSON file with data was retrieved, but something went different than expected
        spec.exit_code(
            302,
            "WARNING_OUTPUT_ZIP_MISSING",
            message="Experiment did not produce a ZIP file with raw data, but the JSON file was parsed."
        )
        spec.exit_code(
            501,
            "WARNING_COMPLETED_ERROR",
            message="The tomato job was marked as completed with an error, but some files were retrieved."
        )
        spec.exit_code(
            502,
            "WARNING_COMPLETED_CANCELLED",
            message="The tomato job was marked as cancelled, but some files were retrieved."
        )

    def prepare_for_submission(self, folder):
        """
        Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files
            needed by the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """

        # if connecting to a Windows PowerShell computer, change the extension of the submit script to '.ps1'
        if self.inputs.code.computer.transport_type == "ssh2win":
            submit_script_filename = self.node.get_option("submit_script_filename")
            if not submit_script_filename.endswith(".ps1"):
                if submit_script_filename.endswith(".sh"):
                    submit_script_filename = submit_script_filename[:-3] + ".ps1"
                else:
                    submit_script_filename(submit_script_filename + ".ps1")
                self.node.backend_entity.set_attribute("submit_script_filename", submit_script_filename)

        # prepare the payload
        TomatoPayload = payload_models[self._INPUT_PAYLOAD_VERSION]

        tomato_dict = self.inputs.control_settings.get_dict()
        if tomato_dict["output"]["prefix"] is None:
            tomato_dict["output"]["prefix"] = self._OUTPUT_FILE_PREFIX
        else:
            self._OUTPUT_FILE_PREFIX = tomato_dict["output"]["prefix"]

        payload = TomatoPayload(
            version=self._INPUT_PAYLOAD_VERSION,
            sample=conversion_map[self._INPUT_PAYLOAD_VERSION]["sample"](self.inputs.battery_sample.get_dict()),
            method=conversion_map[self._INPUT_PAYLOAD_VERSION]["method"](self.inputs.protocol.get_dict()),
            tomato=conversion_map[self._INPUT_PAYLOAD_VERSION]["tomato"](**tomato_dict),
        )

        # HOTFIX: Better to fix beforehand (to-do)
        payload_dict = payload.dict()

        if 'method' in payload_dict:
            new_methods = []

            for old_method in payload_dict['method']:
                new_method = dict(old_method)
                if 'loop' == new_method['technique']:
                    new_method['goto'] = new_method['goto'] - 1
                new_methods.append(new_method)

            payload_dict['method'] = new_methods
        # END HOTFIX

        with folder.open(self.options.input_filename, "w", encoding="utf8") as handle:
            handle.write(yaml.dump(payload_dict))

        codeinfo = datastructures.CodeInfo()

        # the calculation code should be 'ketchup', so we add the following `cmdline_params`
        # in order to submit the payload to tomato
        codeinfo.cmdline_params = [
            "submit",
            "--jobname",
            "$JOB_TITLE",
            self._INPUT_PAYLOAD_YAML_FILE,
        ]
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.withmpi = self.inputs.metadata.options.withmpi

        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.local_copy_list = []
        calcinfo.retrieve_list = [f"{self._OUTPUT_FILE_PREFIX}.json"]
        calcinfo.retrieve_temporary_list = [f"{self._OUTPUT_FILE_PREFIX}.zip"]

        return calcinfo

    def parse_scheduler_output(self, retrieved: Node) -> Optional[ExitCode]:
        """Parse the output of the scheduler if that functionality has been implemented for the plugin."""
        # Tomato scheduler does not return any output that should be parsed
        return None



from aiida import orm
from aiida.engine import ToContext, WorkChain, append_, while_
from aiida.plugins import CalculationFactory, DataFactory

CyclerCalcjob = CalculationFactory('aurora.cycler')
CyclingSpecsData = DataFactory('aurora.cyclingspecs')
BatterySampleData = DataFactory('aurora.batterysample')
TomatoSettingsData = DataFactory('aurora.tomatosettings')


def validate_inputs(inputs, ctx=None):
    """Validate the inputs of the entire input namespace."""
    error_message = ''

    reference_keys = set(inputs['protocol_order'])
    for namekey in inputs['control_settings'].keys():

        if namekey not in reference_keys:
            error_message += f'namekey {namekey} missing in protocols\n'
            continue

        reference_keys.remove(namekey)

    for remaining_key in reference_keys:
        error_message += f'protocol {remaining_key} has no settings\n'

    if len(error_message) > 0:
        return error_message


class CyclingSequenceWorkChain(WorkChain):
    """This workflow represents a process containing a variable number of steps."""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""

        super().define(spec)

        spec.input(
            "battery_sample",
            valid_type=BatterySampleData,
            help="Battery sample to be used.",
        )

        spec.input(
            "tomato_code",
            valid_type=orm.Code,
            help="Tomato code to use.",
        )

        spec.input(
            "protocol_order",
            valid_type=orm.List,
            help="List of experiment protocol names in order of execution.",
        )

        spec.input_namespace(
            "protocols",
            dynamic=True,
            valid_type=CyclingSpecsData,
            help="List of experiment specifications.",
        )

        spec.input_namespace(
            "control_settings",
            dynamic=True,
            valid_type=TomatoSettingsData,
            help="List of experiment control settings.",
        )

        spec.input_namespace(
            "monitor_settings",
            dynamic=True,
            valid_type=orm.Dict,
            help="Dictionary of battery experiment monitor settings.",
        )

        spec.output_namespace(
            "results",
            dynamic=True,
            valid_type=orm.ArrayData,
            help="Results of each step by key.",
        )

        spec.inputs.validator = validate_inputs

        spec.outline(
            cls.setup_workload,
            while_(cls.has_steps_remaining)(
                cls.run_cycling_step,
                cls.inspect_cycling_step,
            ),
            cls.gather_results,
        )

        spec.exit_code(
            401,
            'ERROR_IN_CYCLING_STEP',
            message='One of the steps of CyclingSequenceWorkChain failed',
        )

    def setup_workload(self):
        """Create an ordered list of protocol step names."""
        self.worksteps_keynames = list(self.inputs["protocol_order"])

    def has_steps_remaining(self):
        """Check if there is any remaining step."""
        return len(self.worksteps_keynames) > 0

    def inspect_cycling_step(self):
        """Verify that the last cycling step finished successfully."""
        last_subprocess = self.ctx.subprocesses[-1]

        if not last_subprocess.is_finished_ok:
            pkid = last_subprocess.pk
            stat = last_subprocess.exit_status
            self.report(f'Cycling substep <pk={pkid}> failed with exit status {stat}')
            return self.exit_codes.ERROR_IN_CYCLING_STEP

    def run_cycling_step(self):
        """Run the next cycling step."""
        current_keyname = self.worksteps_keynames.pop(0)

        inputs = {
            'code': self.inputs.tomato_code,
            'battery_sample': self.inputs.battery_sample,
            'protocol': self.inputs.protocols[current_keyname],
            'control_settings': self.inputs.control_settings[current_keyname],
        }

        has_monitors = current_keyname in self.inputs.monitor_settings

        if has_monitors:
            inputs['monitors'] = self.inputs.monitor_settings[current_keyname]

        running = self.submit(CyclerCalcjob, **inputs)
        sample_name = self.inputs.battery_sample.attributes["metadata"]["name"]
        running.label = f"{current_keyname} | {sample_name}"

        if has_monitors:
            running.set_extra('monitored', True)
        else:
            running.set_extra('monitored', False)

        orm.load_group("CalcJobs").add_nodes(running)

        self.report(f'launching CyclerCalcjob<{running.pk}>')
        return ToContext(subprocesses=append_(running))

    def gather_results(self):
        """Gather the results from all cycling steps."""
        keynames = list(self.inputs['protocols'].keys())
        if len(self.ctx.subprocesses) != len(keynames):
            raise RuntimeError('Problem with subprocess!')

        multiple_results = {}
        for keyname in keynames:
            current_subprocess = self.ctx.subprocesses.pop(0)
            if 'results' not in current_subprocess.outputs:
                continue
            multiple_results[keyname] = current_subprocess.outputs.results

        self.out('results', dict(multiple_results))



"""
Tools to convert aurora schemas to dgbowl-schemas
"""

from dgbowl_schemas.tomato.payload_0_1 import Payload as Payload_0p1
from dgbowl_schemas.tomato.payload_0_1.method import Method as Method_0p1
from dgbowl_schemas.tomato.payload_0_1.sample import Sample as Sample_0p1
from dgbowl_schemas.tomato.payload_0_1.tomato import Tomato as Tomato_0p1
from dgbowl_schemas.tomato.payload_0_2 import Payload as Payload_0p2
from dgbowl_schemas.tomato.payload_0_2.method import Method as Method_0p2
from dgbowl_schemas.tomato.payload_0_2.sample import Sample as Sample_0p2
from dgbowl_schemas.tomato.payload_0_2.tomato import Tomato as Tomato_0p2

from .converters.method import electrochemsequence_to_method_list_0
from .converters.sample import batterysample_to_sample_0

payload_models = {
    "0.2": Payload_0p2,
    "0.1": Payload_0p1,
}

conversion_map = {
    "0.1": {
        "method": lambda obj: electrochemsequence_to_method_list_0(obj, Method_0p1),
        "sample": lambda obj: batterysample_to_sample_0(obj, Sample_0p1),
        "tomato": Tomato_0p1,
    },
    "0.2": {
        "method": lambda obj: electrochemsequence_to_method_list_0(obj, Method_0p2),
        "sample": lambda obj: batterysample_to_sample_0(obj, Sample_0p2),
        "tomato": Tomato_0p2,
    },
}



####################################################################################################
"""Script to use as a template for the submission of the cycling sequence workchain."""
####################################################################################################
import datetime

from aiida import orm
from aiida.engine import calcfunction, submit
from aiida.plugins import DataFactory, WorkflowFactory

CyclingSpecsData = DataFactory('aurora.cyclingspecs')
BatterySampleData = DataFactory('aurora.batterysample')
TomatoSettingsData = DataFactory('aurora.tomatosettings')


@calcfunction
def generate_test_inputs():
    """Generate the inputs."""
    nodes_dict = {}

    nodes_dict['sample'] = BatterySampleData({
        'manufacturer':
        'fake_maufacturer',
        'composition':
        dict(description='C|E|A'),
        'form_factor':
        'fake_form',
        'capacity':
        dict(nominal=1.0, units='Ah'),
        'battery_id':
        666,
        'metadata':
        dict(
            name='commercial-10',
            creation_datetime=datetime.datetime.now(tz=datetime.timezone.utc),
            creation_process='This is a fake battery for testing purposes.'
        )
    })

    BASELINE_SPECS = {
        'method': [{
            'name': 'DUMMY_SEQUENTIAL_1',
            'device': 'worker',
            'technique': 'sequential',
            'parameters': {
                'time': {
                    'label': 'Time:',
                    'units': 's',
                    'value': 60.0,
                    'required': True,
                    'description': '',
                    'default_value': 100.0
                },
                'delay': {
                    'label': 'Delay:',
                    'units': 's',
                    'value': 1.0,
                    'required': True,
                    'description': '',
                    'default_value': 1.0
                }
            },
            'short_name': 'DUMMY_SEQUENTIAL',
            'description': 'Dummy worker - sequential numbers'
        }]
    }

    step1_tech = dict(BASELINE_SPECS)
    nodes_dict['step1_tech'] = CyclingSpecsData(step1_tech)
    step2_tech = dict(BASELINE_SPECS)
    step2_tech['method'][0]['parameters']['time']['value'] = 30.0
    nodes_dict['step2_tech'] = CyclingSpecsData(step2_tech)

    BASELINE_SETTINGS = {
        "output": {
            "path": None,
            "prefix": None
        },
        "snapshot": None,
        "verbosity": "INFO",
        "unlock_when_done": True,
    }

    nodes_dict['step1_setting'] = TomatoSettingsData(BASELINE_SETTINGS)
    nodes_dict['step2_setting'] = TomatoSettingsData(BASELINE_SETTINGS)

    return nodes_dict


####################################################################################################
# ACTUAL SUBMISSION

WorkflowClass = WorkflowFactory('aurora.cycling_sequence')
workflow_builder = WorkflowClass.get_builder()
workflow_inputs = generate_test_inputs()

workflow_builder.tomato_code = orm.load_code('ketchup-0.2rc2@localhost-tomato')

workflow_builder.battery_sample = workflow_inputs['sample']
workflow_builder.protocols = {
    'step1': workflow_inputs['step1_tech'],
    'step2': workflow_inputs['step2_tech'],
}
workflow_builder.control_settings = {
    'step1': workflow_inputs['step1_setting'],
    'step2': workflow_inputs['step2_setting'],
}

workflow_node = submit(workflow_builder)

####################################################################################################



import numpy as np
from scipy.integrate import cumtrapz

from aiida.orm import ArrayData


def get_data_from_raw(jsdata) -> dict:
    "Extract raw data from json file."

    if not isinstance(jsdata, dict):
        raise TypeError('jsdata should be a dictionary')

    if len(jsdata["steps"]) > 1:
        raise NotImplementedError('Analysis of multiple steps is not implemented.')

    raw_data = jsdata["steps"][0]["data"]

    # extract raw data
    t = np.array([ts["uts"] for ts in raw_data]) - raw_data[0]["uts"]
    Ewe = np.array([ts["raw"]["Ewe"]["n"] for ts in raw_data])
    I = np.array([ts["raw"]["I"]["n"] for ts in raw_data])

    return post_process_data(t, Ewe, I)


def get_data_from_results(array_node) -> dict:
    "Extract data from parsed ArrayData node."

    if not isinstance(array_node, ArrayData):
        raise TypeError('array_node should be an ArrayData')

    # collect data
    t = array_node.get_array('step0_uts')
    t -= t[0]
    Ewe = array_node.get_array('step0_Ewe_n')
    I = array_node.get_array('step0_I_n')

    return post_process_data(t, Ewe, I)


def post_process_data(t: np.ndarray, Ewe: np.ndarray, I: np.ndarray) -> dict:
    """docstring"""

    # find half-cycle markers
    # add last point if not already a marker
    idx = np.where(np.diff(np.sign(I)) != 0)[0]
    if (final := len(I) - 1) not in idx:
        idx = np.append(idx, final)

    # integrate and store charge and discharge currents
    Qc, Qd = [], []
    for ii in range(len(idx) - 1):
        i0, ie = idx[ii], idx[ii + 1]
        if ie - i0 < 10:
            continue
        q = np.trapz(I[i0:ie], t[i0:ie])
        if q > 0:
            Qc.append(q)
        else:
            Qd.append(abs(q))

    return {
        'time': t,
        'Ewe': Ewe,
        'I': I,
        'cn': len(Qd),
        'time-cycles': t[idx[2::2]],
        'Q': cumtrapz(I, t, axis=0, initial=0) / 3.6,
        'Qd': np.array(Qd) / 3.6,
        'Qc': np.array(Qc) / 3.6,
    }



from itertools import groupby
from logging import LoggerAdapter
from typing import Optional

from aiida.common.log import AIIDA_LOGGER, LOG_LEVEL_REPORT

from .parsers import get_data_from_raw


class Analyzer:
    """Base class for all analyzers.

    Attributes
    ==========
    `logger` : `Union[AiidaLoggerType, LoggerAdapter]`
        The associated logger.
    """

    logger = AIIDA_LOGGER.getChild("monitor")

    def set_logger(self, logger: LoggerAdapter) -> None:
        """Set the analyzer logger.

        Parameters
        ----------
        `logger` : `LoggerAdapter`
            The logger of the analyzed calculation node.
        """
        self.logger = logger

    def analyze(self, snapshot: dict) -> Optional[str]:
        """Analyze the experiment snapshot against a condition.

        Condition is defined in subclass analyzers.

        Parameters
        ----------
        `snapshot` : `dict`
            The loaded snapshot dictionary.

        Returns
        -------
        `Optional[str]`
            A string if a defined condition has been met,
            `None` otherwise.
        """
        raise NotImplementedError


class CapacityAnalyzer(Analyzer):
    """A battery capacity analyzer.

    Attributes
    ==========
    `check_type` : `str`
        The half-cycle to analyze (charge/discharge),
        `"discharge_capacity"` by default.
    `threshold` : `float`
        The capacity threshold in percent, `0.8` by default.
    `consecutive_cycles` : `int`
        The number of required below-threshold consecutive cycles,
        `2` by default
    """

    def __init__(
        self,
        check_type="discharge_capacity",
        threshold=0.8,
        consecutive_cycles=2,
    ) -> None:
        """`CapacityAnalyzer` constructor.

        Parameters
        ----------
        `check_type` : `str`
            The half-cycle to analyze,
            `"discharge_capacity"` by default.
        `threshold` : `float`
            The capacity threshold in percent, `0.8` by default.
        `consecutive_cycles` : `int`
            The number of required consecutive cycles,
            `2` by default.

        Raises
        ------
        `TypeError`
            If `check_type` is not supported.
        """

        if check_type not in {"discharge_capacity", "charge_capacity"}:
            raise TypeError(f"{check_type=} not supported")

        self.threshold = threshold
        self.consecutive = consecutive_cycles
        self.is_discharge = check_type == "discharge_capacity"

    def analyze(self, snapshot: dict) -> Optional[str]:
        """Analyze the snapshot.

        Check if capacity has fallen below threshold for required
        consecutive cycles.

        Parameters
        ----------
        `snapshot` : `dict`
            The loaded snapshot dictionary.

        Returns
        -------
        `Optional[str]`
            If condition is met, an exit message, `None` otherwise.
        """
        self.capacities = self._get_capacities(snapshot)
        self.cycles = len(self.capacities)
        return None if self.cycles < 1 else self._check_capacity()

    ###########
    # PRIVATE #
    ###########

    def _get_capacities(self, snapshot: dict):
        """Post-process the snapshot to extract capacities.

        Parameters
        ----------
        `snapshot` : `dict`
            The loaded snapshot dictionary.

        Returns
        -------
        `_type_`
            A `numpy` array of capacities (in mAh), or empty list
            if failed to process snapshot.
        """
        try:
            data = get_data_from_raw(snapshot)
            return data['Qd'] if self.is_discharge else data['Qc']
        except KeyError as err:
            self.logger.error(f"missing '{str(err)}' in snapshot")
        return []

    def _check_capacity(self) -> Optional[str]:
        """Check if capacity has fallen below threshold for required
        consecutive cycles.

        Returns
        -------
        `Optional[str]`
            If condition is met, an exit message, `None` otherwise.
        """

        n = self.cycles
        Qs = self.capacities[0]
        Q = self.capacities[-1]
        Qt = self.threshold * Qs

        message = f"cycle #{n} : {Q = :.2f} mAh ({Q / Qs * 100:.1f}%)"

        if Q < Qt:
            message += f" : {(Qt - Q) / Qt * 100:.1f}% below threshold"

        self.logger.log(LOG_LEVEL_REPORT, message)

        below_threshold = self._count_cycles_below_threshold()
        if below_threshold >= self.consecutive:
            return f"Capacity below threshold ({Qt:.2f} mAh) " \
                   f"for {below_threshold} cycles!"

        return None

    def _count_cycles_below_threshold(self) -> int:
        """Count the number of consecutive cycles below threshold.

        Returns
        -------
        `int`
            The number of consecutive cycles below threshold.
        """
        Qt = self.threshold * self.capacities[0]
        return next(
            (
                len(list(group))  # cycle-count of first below-threshold group
                for below, group in groupby(self.capacities < Qt)
                if below
            ),
            0,
        )
